from os import system
from sys import platform
from typing import Iterable


class ColorCodes:
    """
    A class representation of the color codes used by `samutil.Formatter`
    """
    CYAN = "\033[96m"
    WHITE = "\033[97m"
    MAGENTA = "\033[95m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    # Returns terminal color to normal
    END = "\033[0m"


class ColoredText(ColorCodes):
    def __init__(self, code: str, val: str):
        self.code = code
        self.val = val

    def __add__(self, rhs: str) -> str:
        """
        Defines what to do in the case of a concatenation like ColoredText + "Hello".
        """
        if isinstance(rhs, str):
            return self.code + self.val + self.END + rhs
        elif isinstance(rhs, ColoredText):
            return self.code + self.val + self.END + rhs.code + rhs.val
        else:
            raise TypeError(
                f"Operand '+' is not supported between types {type(self)} and {type(rhs)}."
            )

    def __radd__(self, lhs: str) -> str:
        """
        Defines what to do in the case of a concatenation like "Hello" + ColoredText.
        """
        if isinstance(lhs, str):
            return lhs + f"{self.code}{self.val}{self.END}"

    def __repr__(self) -> str:
        """
        Format the ColoredText object into a string
        """
        return f"{self.code}{self.val}"

    def __str__(self) -> str:
        """
        Format the ColoredText object into a string
        """
        return f"{self.code}{self.val}{self.END}"


class Formatter:
    """
    Class with methods for changing the color of text.
    """

    # Required to make ANSI codes work on Windows
    if platform == "win32":
        system("color")

    @staticmethod
    def _concat(strings: Iterable[str], code: str = "", sep: str = " ") -> str:
        """
        Helper function for concatenating multiple arguments passed
        to any Formatter methods properly.
        """
        if len(strings) == 1:
            sep = ""

        combined_strings = ""
        for string in strings:

            if not (hasattr(string, "__str__") or hasattr(string, "__repr__")):
                raise TypeError(
                    f"Value '{string}' of type '{type(string)}' doesn't have __str__ or __repr__ method."
                )

            string = str(string)

            if isinstance(string, ColoredText):
                combined_strings += string + sep
            else:
                # This is to ensure that nested formatter calls work as intended
                combined_strings += code + string + sep

        return combined_strings

    @staticmethod
    def info(*strings: str, sep=" ") -> str:
        """
        Return `strings` as 1 Cyan colored string, separated by `sep`
        """
        text = Formatter._concat(strings, ColorCodes.CYAN, sep)
        return ColoredText(ColorCodes.CYAN, text)

    @staticmethod
    def success(*strings: str, sep=" ") -> str:
        """
        Return `strings` as 1 Green colored string, separated by `sep`
        """
        text = Formatter._concat(strings, ColorCodes.GREEN, sep)
        return ColoredText(ColorCodes.GREEN, text)

    @staticmethod
    def warning(*strings: str, sep=" ") -> str:
        """
        Return `strings` as 1 Yellow colored string, separated by `sep`
        """

        text = Formatter._concat(strings, ColorCodes.YELLOW, sep)
        return ColoredText(ColorCodes.YELLOW, text)

    @staticmethod
    def error(*strings: str, sep=" ") -> str:
        """
        Return `strings` as 1 Red colored string, separated by `sep`
        """
        text = Formatter._concat(strings, ColorCodes.RED, sep)
        return ColoredText(ColorCodes.RED, text)

    @staticmethod
    def magenta(*strings: str, sep=" ") -> str:
        """
        Return `strings` as 1 Magenta colored string, separated by `sep`
        """
        text = Formatter._concat(strings, ColorCodes.MAGENTA, sep)
        return ColoredText(ColorCodes.MAGENTA, text)

    @staticmethod
    def bold(*strings: str, sep=" ") -> str:
        """
        Return `strings` as 1 Bold string, separated by `sep`
        """
        text = Formatter._concat(strings, ColorCodes.BOLD, sep)
        return ColoredText(ColorCodes.BOLD, text)

    @staticmethod
    def underline(*strings: str, sep=" ") -> str:
        """
        Return `strings` as 1 Underlined string, separated by `sep`
        """
        text = Formatter._concat(strings, ColorCodes.UNDERLINE, sep)
        return ColoredText(ColorCodes.UNDERLINE, text)



import inspect
from types import FunctionType
from typing import Callable

from samutil.formatting import Formatter as f

from .comparisons import BaseComparison, EqualTo
from .core import UnitTest
from .utils import make_lazy_run_test


def expect(*args):
    """
    Specify the expected value that should be returned by the test case.
    Must always be used directly underneath a `@case`.
    """
    if len(args) != 1:
        raise ValueError(
            f.error(
                "@expect only takes 1 argument, the value returned by one of the Comparisons methods, or a literal"
            )
        )

    def deco(func: Callable) -> Callable:
        func._is_class = inspect.isclass(func)
        comparison = args[0]
        if not hasattr(func, "_tests"):
            func._tests = [[]]

        if not isinstance(comparison, BaseComparison):
            comparison = EqualTo(comparison)

        index = 0
        if hasattr(func, "_test_index"):
            index = func._test_index

        try:
            func._tests[index].append(comparison)
        except IndexError:
            func._tests.append([comparison])

        return func

    return deco


def case(*args, **kwargs) -> Callable:
    """
    Decorator which wraps a function and automatically creates a test case with it.
    """

    def deco(func: Callable) -> Callable:
        if not hasattr(func, "_tests"):
            raise ValueError(
                f.error(f"@case called on '{func.__name__}' without @expect underneath")
            )

        index = 0
        if hasattr(func, "_test_index"):
            index = func._test_index

        comparison = func._tests[index][-1]
        del func._tests[index][-1]
        func._tests[index].append(
            make_lazy_run_test(*args, comparison=comparison, **kwargs)
        )
        return func

    return deco


def test(*args):
    """
    Create a new test suite for a function which can be run using `samutil test <file_with_function_declaration>`
    """
    if len(args) > 1:
        raise ValueError(
            f.error(
                f"@test accepts 1 argument: `name` which is the name of the unit test. Got {len(args)} arguments: {', '.join(args)}"
            )
        )

    def deco(func: Callable):
        try:
            name = str(args[0])
            if isinstance(name, FunctionType):
                name = func.__name__
        except (IndexError, ValueError):
            name = func.__name__

        if hasattr(func, "_is_class") and func._is_class:
            raise TypeError(
                f.error(
                    "@test can't be used on a class. For testing classes, use @testmethod"
                )
            )

        test = UnitTest(func)
        test.describe(name)

        # Used for grouping like tests together
        if not hasattr(func, "_test_index"):
            func._test_index = 1
        else:
            func._test_index += 1

        # Used for grouping callbacks that run tests
        this_suite = 0
        if hasattr(func, "_run_tests"):
            this_suite = len(func._run_tests)
        else:
            func._run_tests = []

        if not hasattr(func, "_tests"):
            raise ValueError(
                f.error(
                    f"@test called on '{func.__name__}' without any @case statements underneath"
                )
            )

        def run_tests(fn):
            for case in fn._tests[this_suite][::-1]:
                case(test)()

        func._run_tests.append(run_tests)

        return func

    if isinstance(args[0], FunctionType):
        return deco(args[0])
    else:
        return deco


def testmethod(*args):
    """
    Create a new test suite for a class method which can be run using `samutil test <file_with_class_declaration>`
    """
    arg_len = len(args)
    if arg_len > 2:
        raise ValueError(
            f.error(
                f"@testmethod accepts 2 arguments: 'methodname' and 'testname'. Got {len(args)} arguments: {', '.join(args)}"
            )
        )
    elif arg_len == 0:
        raise ValueError(
            f.error(
                "@testmethod must be called without arguments. The first argument must be the name of the method."
            )
        )

    def deco(func: Callable):
        methodname = args[0]

        if arg_len == 2:
            testname = args[1]
        else:
            testname = func.__name__

        if hasattr(func, "_is_class") and func._is_class:
            method = getattr(func, methodname)
            method.__dict__.update(func.__dict__)
            method._parent = func
            test = UnitTest(method, testname)

        this_suite = 0

        if not hasattr(func, "_test_index"):
            func._test_index = 1
        else:
            func._test_index += 1

        has_tests = hasattr(func, "_run_tests")

        if has_tests:
            this_suite = len(func._run_tests)

        if not hasattr(func, "_tests"):
            raise ValueError(
                f.error(
                    f"@test called on '{func.__name__}' without any @case statements underneath"
                )
            )

        def run_tests(fn):
            print("\n" + f.underline(testname + "\n"))

            for case in fn._tests[this_suite][::-1]:
                case(test)()

        if not has_tests:
            func._run_tests = [run_tests]
        else:
            func._run_tests.append(run_tests)

        return func

    return deco



import importlib.util
import os
from types import FunctionType, ModuleType
from typing import Callable, Tuple

import click
from click.types import Path
from samutil.formatting import Formatter as f

from .types import TestSubject, Value


def call_if_callable(obj: object, *args, **kwargs) -> Tuple[Value, float]:
    """
    Take an object and any amout of args, and call the object with the args
    if it can be called, with performance metrics.
    """
    if callable(obj):
        if getattr(obj, "_is_class", False) and (not isinstance(obj, FunctionType)):
            response = obj(obj._parent(), *args, **kwargs)
        else:
            response = obj(*args, **kwargs)
        return response
    else:
        return obj, 0


def output_case_args(test_subject: TestSubject, *args, **kwargs):
    """
    Take a test subject and all of the test's arguments and output it nicely.
    """
    formatted_kwargs = [f"{k}={v}" for (k, v) in kwargs.items()]
    args = tuple(map(str, args))
    formatted_args = ", ".join([*args, *formatted_kwargs])
    formatted_output = f.bold(test_subject.__name__ + "(" + formatted_args + ")")
    print(f.info("  RUNS", f.bold(formatted_output), "\b:"))


def lazy_run_test(test, *args, comparison, **kwargs) -> Callable:
    case = test.with_args(*args, **kwargs)

    return lambda: case.should_be(comparison)


def make_lazy_run_test(*args, comparison, **kwargs):
    return lambda test: lazy_run_test(test, *args, comparison=comparison, **kwargs)


def import_file(filename: str, search: bool) -> ModuleType:
    """
    Import the contents of a file and return it as a module
    """

    spec = importlib.util.spec_from_file_location("mod", filename)
    mod = importlib.util.module_from_spec(spec)
    try:
        spec.loader.exec_module(mod)
    except Exception as e:
        if not search:
            print(f.error(f"Got error: '{e}' when executing '{filename}', skipping..."))
        return

    return mod


def module_funcs(module: ModuleType):
    for func in module.__dict__.items():
        yield func


def test_file(filename: str, search: bool):
    """
    Dynamically import a python file and check for functions declared with @case and @test
    """

    module = import_file(filename, search=search)
    if module:
        for func in module_funcs(module):
            # func is a tuple in the form ('func_name', actual_func)
            func = func[1]
            if callable(func) and hasattr(func, "_run_tests"):
                print(f.bold("\nFile:", filename))
                for test in func._run_tests[::-1]:
                    test(func)

    del module


def test_test_file(filename: str):
    with open(filename) as f:
        code = compile(f.read(), filename, "exec")
        exec(code)


def test_dir(dir: Path, search: bool):
    ignore_dirs = ["__pycache__", "venv", "env", "virtualenv", "build", "dist"]
    dirname = click.format_filename(dir)
    for file in os.listdir(dirname):
        filename = os.path.join(dirname, file)

        if os.path.isdir(filename):
            if filename not in ignore_dirs:
                test_dir(filename, search=search)
        elif os.path.isfile(filename):
            if filename.endswith(".py"):
                if filename.endswith(".test.py"):
                    test_test_file(filename)
                else:
                    test_file(filename, search=search)



from samutil.formatting import Formatter as f

from .comparisons import ComparisonRunner
from .types import TestSubject


class UnitTest:
    def __init__(self, test_subject: TestSubject):
        """
        Instantiate a new UnitTest with a value or callback which is the
        subject of the test.
        """
        self._test_subject = test_subject
        is_callable = callable(self._test_subject)

        if not is_callable:
            if not (
                hasattr(self._test_subject, "__str__")
                or hasattr(self._test_subject, "__repr__")
            ):
                raise ValueError(
                    f"UnitTest value '{self._test_subject} of type '{type(self._test_subject)}' has no __str__ or __repr__ method"
                )
            name = str(self._test_subject)
            self._test_subject = lambda: self._test_subject
        else:
            name = self._test_subject.__name__

        self._test_subject._name = name

    def describe(self, testname: str):
        self._test_subject._name = testname

    def value(self):
        """
        Create a new test case using the value passed to `UnitTest` when it was instantiated.
        """
        print("\n" + f.underline(self._test_subject._name + "\n"))
        return ComparisonRunner(self._test_subject)

    def with_args(self, *args, **kwargs):
        """
        Create a new test case, which calls the callable passed to the `UnitTest` when
        it was instantiated using the the args passed to this method.
        """
        print("\n" + f.underline(self._test_subject._name + "\n"))
        return ComparisonRunner(self._test_subject, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        """
        Create a new test case, which calls the callable passed to the `UnitTest` when
        it was instantiated using the the args passed to this method.
        """
        return self.with_args(*args, **kwargs)

    def string(self):
        """
        Run a test using the return value of the __str__ method defined on the
        test subject.
        """
        print("\n" + f.underline(self._test_subject._name + "\n"))
        return ComparisonRunner(str(self._test_subject))

    def type(self):
        print("\n" + f.underline(self._test_subject._name + "\n"))
        return ComparisonRunner(type(self._test_subject))

    def method(self, methodname: str):
        """
        Run a test on a specific method of the test subject
        """
        method = getattr(self._test_subject, methodname, None)
        if not method:
            raise AttributeError(
                f.error(
                    f"Method '{methodname}' does not exist on object {self._test_subject.__name__}"
                )
            )

        test = UnitTest(method)
        test.describe(self._test_subject._name)
        return test

    def __str__(self):
        return self._test_subject._name



from os import getcwd, path

import click

from .formatting import Formatter as f
from .generation.core import generate_key
from .testing.utils import test_dir, test_file, test_test_file


@click.group("samutil")
@click.version_option("0.0.71")
def main():
    """Samutil Python CLI"""
    pass


@main.command("test")
@click.argument("filenames", nargs=-1, required=False)
def test(filenames: tuple[click.Path]):
    if len(filenames) == 0:
        test_dir(getcwd(), search=True)
    else:
        if filenames[0] == ".":
            print(f.error("To test an entire directory, use '*'"))
            return
        for file in filenames:
            filename = click.format_filename(file)
            ignore_dirs = ["__pycache__", "venv", "env", "virtualenv", "build", "dist"]
            if path.isdir(filename):
                if filename not in ignore_dirs:
                    test_dir(filename, search=False)
            elif path.isfile(filename) and filename.endswith(".py"):
                if filename.endswith(".test.py"):
                    test_test_file(filename)
                else:
                    test_file(filename, search=False)
            else:
                print(
                    f.warning(
                        "WARNING: Skipping",
                        f.underline(f.bold(str(file))),
                        "as it is not a directory or valid test file.",
                    )
                )


@main.command("key")
@click.option("-l", "--length", type=int, default=6)
@click.option("-a", "--amount", type=int, default=1)
@click.option("-o", "--out", type=str, default=None)
def key(length: int = 6, amount: int = 1, out: click.Path = None):
    tokens = []
    for _ in range(amount):
        tokens.append(generate_key(length))

    if out:
        with open(out, "w+") as o:
            o.write("\n".join(tokens))
            print(f.success("Wrote tokens to", f.bold(out)))
    else:
        print("\n" + "\n".join(tokens) + "\n")


if __name__ == "__main__":
    main()



from time import perf_counter

from samutil.formatting import Formatter as f
from sigfig import round

from .types import TestSubject, Value
from .utils import call_if_callable, output_case_args


class BaseComparison:
    result = None
    time_taken = 0
    same_type = True
    passed = False
    operator = "?"
    negated = "?"
    _not = False

    def __init__(self, expected: Value):
        self.expected = expected

    def compare(self, expected: Value, result: Value) -> bool:
        """
        If compare has not been implemented by a child class, raise an error
        """
        raise AttributeError(
            f.error(
                f"Custom comparison implementation '{self.__name__}' must implement a 'compare' method."
            )
        )


class EqualTo(BaseComparison):
    operator = "=="
    negated = "!="

    def compare(self, result, expected):
        return result == expected


def Not(comp: BaseComparison):
    """
    Negate the value that is being expected
    """
    if isinstance(comp, BaseComparison):
        comp._not = True
    else:
        comp = EqualTo(comp)
        comp._not = True
    return comp


class ListEqual(EqualTo):
    def compare(self, result, expected):
        if not isinstance(result, (list, tuple)):
            return False

        for result_val, expected_val in zip(result, expected):
            if result_val != expected_val:
                return False

        return True


class DictEqual(EqualTo):
    def compare(self, result: dict, expected: dict):
        if not isinstance(result, dict):
            return False

        for result_val, expect_val in zip(result.values(), expected.values()):
            if isinstance(result_val, dict):
                if not self.compare(result_val, expect_val):
                    return False
            elif result_val != expect_val:
                return False

        return True


class HasType(BaseComparison):
    operator = "instance of"
    negated = "not instance of"

    def compare(self, result, expected):
        return isinstance(result, type(expected))


class LessThan(BaseComparison):
    operator = "<"
    negated = "!<"

    def compare(self, result, expected):
        return result < expected


class LessThanOrEqualTo(BaseComparison):
    operator = "<="
    negated = "!<="

    def compare(self, result, expected):
        return result <= expected


class GreaterThan(BaseComparison):
    operator = ">"
    negated = "!>"

    def compare(self, result, expected):
        return result > expected


class GreaterThanOrEqualTo(BaseComparison):
    operator = ">="
    negated = "!>="

    def compare(self, result, expected):
        return result >= expected


class ToRaise(BaseComparison):
    operator = "instance of"
    negated = "not instance of"

    def compare(self, result, expected):
        return isinstance(result, expected)


class ComparisonRunner:
    def __init__(self, test_subject: TestSubject, *args, **kwargs):
        self._test_subject = test_subject
        self._args = args
        self._kwargs = kwargs

        output_case_args(test_subject, *args, **kwargs)

    def _run(self, comparison: BaseComparison):
        t1 = perf_counter()

        # Only catch errors if the test is checking for an exception
        if isinstance(comparison, ToRaise):
            try:
                comparison.result = call_if_callable(
                    self._test_subject, *self._args, **self._kwargs
                )
            except Exception as e:
                comparison.result = e
        else:
            comparison.result = call_if_callable(
                self._test_subject, *self._args, **self._kwargs
            )

        t2 = perf_counter()
        comparison.time_taken = t2 - t1

        comparison.passed = comparison.compare(comparison.result, comparison.expected)

        if comparison._not:
            comparison.passed = not comparison.passed
            comparison.operator, comparison.negated = (
                comparison.negated,
                comparison.operator,
            )

        # Check to see if result has same type as expected value,
        # If it doesn't, try to cast it.
        should_be_type = type(comparison.expected)
        if not isinstance(comparison.result, should_be_type):
            comparison.same_type = False

        return comparison

    def _parse(self, comparison: BaseComparison):
        time_unit = "seconds"
        time_taken = comparison.time_taken

        if time_taken > 0 and time_taken <= 0.01:
            time_taken = time_taken * 1000  # Convert to milliseconds
            time_unit = "milliseconds"

            if time_taken <= 0.01:
                time_taken = time_taken * 1000  # Convert to microseconds
                time_unit = "microseconds"

        time_taken = round(time_taken, sigfigs=3)

        if not comparison.passed:
            print(f.error("    - FAIL -"))

            print()

            print(
                f.bold(f.error("    Received")),
                comparison.negated,
                f.success(f.bold("Expected")),
            )
            print(
                f.bold("    " + f.error(comparison.result)),
                comparison.negated,
                f.success(f.bold(comparison.expected)),
            )

            if not comparison.same_type:
                print(
                    f.warning(
                        f"\n    Warning: expected and received values had different types."
                    )
                )
                print(
                    f.success(
                        f"      Expected: {type(comparison.expected).__name__} ({comparison.expected})"
                    )
                )
                print(
                    f.error(
                        f"      Received: {type(comparison.result).__name__} ({comparison.result})"
                    )
                )

            time_taken != 0 and print(
                f.magenta(f"\n  Execution time:", f.bold(time_taken, time_unit)) + "\n"
            )
        else:
            print(f.success("    - PASS -"))
            if not comparison.same_type:
                print(
                    f.warning(
                        f"\n    Warning: expected and received values had different types."
                    )
                )
                print(
                    f.success(
                        f"      Expected: {type(comparison.expected).__name__} ({comparison.expected})"
                    )
                )
                print(
                    f.error(
                        f"      Received: {type(comparison.result).__name__} ({comparison.result})\n"
                    )
                )
            time_taken != 0 and print(
                f.magenta(f"  Execution time:", f.bold(time_taken, time_unit)) + "\n"
            )

    def should_equal(self, expected: Value):
        """
        Result of call should be equal to expected
        """
        self._parse(self._run(EqualTo(expected)))

    def should_be_less_than(self, expected: Value):
        """
        Result of call should be less than expected
        """
        self._parse(self._run(LessThan(expected)))

    def should_be_less_or_equal_to(self, expected: Value):
        """
        Result of call should be less than or equal to expected
        """
        self._parse(self._run(LessThanOrEqualTo(expected)))

    def should_be_greater_than(self, expected: Value):
        """
        Result of call should be greater than to expected
        """
        self._parse(self._run(GreaterThan(expected)))

    def should_be_greater_or_equal_to(self, expected: Value):
        """
        Result of call should be greater than or equal to expected
        """
        self._parse(self._run(GreaterThanOrEqualTo(expected)))

    def should_be(self, comparison: BaseComparison):
        """
        Pass a custom comparison class which extends `BaseComparison` for use in unit test.
        """
        if not isinstance(comparison, BaseComparison):
            raise TypeError(
                f.error(
                    "Argument passed to 'should' method must be a class which extends from BaseComparison, or one of the pre-defined comparison objects."
                )
            )

        self._parse(self._run(comparison))



from secrets import choice
from string import ascii_letters, digits, punctuation

KEY_LENGTH = 84
punctuation = punctuation.replace('"', "").replace("'", "")
ALLOWED_CHARACTERS = ascii_letters + digits + punctuation


def generate_key(
    length: int = KEY_LENGTH, allowed_characters: str = ALLOWED_CHARACTERS
) -> str:
    """
    Generate a key using a given `length` from the set of charcters `allowed_characters`
    """
    key = ""
    for _ in range(length):
        key += choice(allowed_characters)

    assert len(key) == length
    return key



from setuptools import find_packages, setup

setup(
    name="samutil",
    description="A set of Python utility packages for developing and maintaining quality software.",
    version="0.0.73",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=["click", "sigfig", "sortedcontainers"],
    python_requires=">=3",
    include_package_data=True,
    py_modules=["cli"],
    entry_points="""
        [console_scripts]
        samutil = samutil.cli:main
    """,
    author="Sam McElligott",
    keyword="testing, unittesting, formatting, secure, generation, utility",
    long_description="README.md",
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/sammce/samutil",
    project_urls={
        "Bug Tracker": "https://github.com/sammce/samutil/issues",
    },
    author_email="sammcelligott@outlook.com",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
)
